import 'dart:async';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:get/get.dart';
import 'package:_private_chat/chat_detail/chat_detail_screen.dart';
import 'package:_private_chat/chat_detail/chat_helper.dart';
import 'package:_private_chat/utils/utils.dart';
import 'package:_private_core/_private_core.dart';

import '../../connection_status/bloc/connection_status_bloc.dart';

part 'chat_list_event.dart';
part 'chat_list_state.dart';

String chatGetFullNameUser(data) =>
    data == null ? '' : data['displayName'] ?? data['username'] ?? '';

String chatGetThumbnailUser(data) => data == null ? '' : data['thumbnail'];

String chatGetUsernameUser(data) => data == null ? '' : data['username'];

class ChatListConversation {
  int timestamp;
  ChatConversationType conversationType;
  String conversationId;
  Map data = {};
  List userIds = [];
  int unread;
  Map<String, dynamic> lastMessage = {};
  bool isEmptyMessages;
  bool isBot;

  ChatListConversation({
    this.timestamp = 0,
    this.isBot = false,
    this.conversationId = '',
    this.conversationType = ChatConversationType.solo,
    this.unread = 0,
    this.isEmptyMessages = false,
  });
}

class ChatListBloc extends Bloc<ChatListEvent, ChatListState> {
  StreamSubscription? _converstationsSubscription;
  final List<StreamSubscription> _msgsSubscription = [];

  late final CollectionReference _col =
      FirebaseFirestore.instance.collection(C_CONVERSATIONS);

  ChatListBloc() : super(ChatListState(conversations: [], usersByKeys: {})) {
    on<ChatListInitEvent>(_onInitial);
    on<ChatListReadItemEvent>(_onReadItemEvent);
    on<ChatListDisposeEvent>(_onDisposeEvent);
    on<ChatListRemoveEvent>(onRemoveEvent);
    on<ChatUpdateListConversationEvent>(_onUpdateConversation);
    on<LoadUserInfoConversationEvent>(_onLoadUserInfo);
  }

  _onUpdateConversation(
      ChatUpdateListConversationEvent event, Emitter<ChatListState> emit) {
    emit(state.update(conversations: event.conversations));
  }

  onRemoveEvent(ChatListRemoveEvent event, Emitter<ChatListState> emit) {
    ChatHelper.removeConversation(event.conversationId);
    state.conversations
        .removeWhere((e) => e.conversationId == event.conversationId);
    emit(state.update());
  }

  _onReadItemEvent(ChatListReadItemEvent event, Emitter<ChatListState> emit) {
    if (event.c != null) {
      if (event.c!.isBot) {
        Get.toNamed(
          ChatDetailScreen.route,
          arguments: ChatDetailParams(
            event.c!.conversationType,
            isBot: true,
          ),
        );
      } else {
        if (event.c!.conversationType == ChatConversationType.solo) {
          String username = event.c!.userIds.first[DB_KEY_USERNAME];
          Map? user = state.usersByKeys[username];
          String fullname = chatGetFullNameUser(user);
          String thumbnail = chatGetThumbnailUser(user);
          Get.toNamed(ChatDetailScreen.route,
              arguments: ChatDetailParams(
                event.c!.conversationType,
                peerId: username,
                peerType: event.detailType ??
                    ((user?['uniqueId']?.startsWith('U') ?? false)
                        ? ChatDetailType.users
                        : ChatDetailType.pros),
                name: fullname,
                thumbnail: thumbnail,
              ));
        } else {
          Get.toNamed(
            ChatDetailScreen.route,
            arguments: ChatDetailParams(
              event.c!.conversationType,
              conversationId: event.c!.conversationId,
              name: event.c!.data[DB_KEY_GROUP_NAME],
              thumbnail: event.c!.data[DB_KEY_THUMBNAIL],
            ),
          );
        }
      }
    } else if (event.username != null) {
      Get.toNamed(
        ChatDetailScreen.route,
        arguments: ChatDetailParams(
          ChatConversationType.solo,
          peerId: event.username,
          peerType: event.detailType,
          name: event.displayName,
          thumbnail: event.thumbnail,
        ),
      );
    }
  }

  _onDisposeEvent(event, Emitter<ChatListState> emit) {
    _converstationsSubscription?.cancel();
    for (var e in _msgsSubscription) {
      e.cancel();
    }
    _msgsSubscription.clear();
  }

  _onLoadUserInfo(
      LoadUserInfoConversationEvent event, Emitter<ChatListState> emit) async {
    var usersByKeys = state.usersByKeys;

    Map<String, dynamic>? data = (await FirebaseFirestore.instance
            .collection(event.type)
            .doc(event.username)
            .get())
        .data();
    usersByKeys[event.username] = data;

    emit(state.update(
      usersByKeys: usersByKeys,
    ));
  }

  _onInitial(ChatListInitEvent event, Emitter<ChatListState> emit) async {
    emit(state.update(isLoading: true));
    QuerySnapshot q = await _col
        .where(DB_KEY_MEMBERS, arrayContains: loggedUsername)
        .orderBy(DB_KEY_LAST_MESSAGE_TIME, descending: true)
        .get();
    List requestUsers = [];
    List<ChatListConversation> cons = [];
    if (currentChatOptions.botEnable) {
      cons.add(ChatListConversation(isBot: true));
    }
    for (var e in _msgsSubscription) {
      e.cancel();
    }
    _msgsSubscription.clear();
    for (var d in q.docs) {
      Map docData = d.data() as Map;
      int timestamp = docData[DB_KEY_MEMBERS_DETAIL].firstWhere(
              (e) => e[DB_KEY_USERNAME] == loggedUsername)[DB_KEY_TIMESTAMP] ??
          0;
      ChatListConversation conversation = ChatListConversation()
        ..timestamp = timestamp
        ..conversationId = d.id
        ..userIds = docData[DB_KEY_MEMBERS_DETAIL]
            .where((e) => e[DB_KEY_USERNAME] != loggedUsername)
            .toList()
        ..conversationType = ChatConversationType.values
            .byName(docData[DB_KEY_CONVERSATION_TYPE])
        ..data = docData;
      requestUsers.addAll(docData[DB_KEY_MEMBERS_DETAIL].toList());

      if (docData[DB_KEY_LAST_MESSAGE_TIME] != null &&
          docData[DB_KEY_LAST_MESSAGE_TIME] != 0) {
        QuerySnapshot query = await _col
            .doc(d.id)
            .collection(C_MESSAGES)
            .where(DB_KEY_TIMESTAMP, isGreaterThan: timestamp)
            .get();
        conversation.unread = query.size;

        //Last message
        DocumentSnapshot docLastMessage = await _col
            .doc(d.id)
            .collection(C_MESSAGES)
            .doc('${docData[DB_KEY_LAST_MESSAGE_TIME]}')
            .get();
        Map<String, dynamic> dataC = {};
        if (docLastMessage.data() != null) {
          dataC = docLastMessage.data() as Map<String, dynamic>;
          dataC[DB_KEY_CONTENT] = ChatHelper.decrypt(
              conversation.data[DB_KEY_ENCRYPT_KEY], dataC[DB_KEY_CONTENT]);
        }
        conversation.lastMessage = dataC;

        cons.add(conversation);
      } else {
        conversation.unread = 0;
        conversation.isEmptyMessages = true;
        conversation.lastMessage = {};
        cons.add(conversation);
      }
    }

    Map usersByKeys = await _getUsersByKeys(requestUsers);
    _getConnectionStatusUsers(requestUsers);

    emit(state.update(
      isLoading: false,
      conversations: cons,
      usersByKeys: usersByKeys,
    ));

    _converstationsSubscription ??= _col
        .where(DB_KEY_MEMBERS, arrayContains: loggedUsername)
        .snapshots()
        .listen((docs) {
      if (docs.docs.length != state.conversations.length) {
        add(ChatListInitEvent());
      }
    });

    for (var d in q.docs) {
      _msgsSubscription.add(_col.doc(d.id).snapshots().listen((snapshot) async {
        Map docData = snapshot.data() as Map;
        int timestamp = docData[DB_KEY_MEMBERS_DETAIL].firstWhere((e) =>
                e[DB_KEY_USERNAME] == loggedUsername)[DB_KEY_TIMESTAMP] ??
            0;
        ChatListConversation conversation = ChatListConversation()
          ..timestamp = timestamp
          ..conversationId = d.id
          ..userIds = docData[DB_KEY_MEMBERS_DETAIL]
              .where((e) => e[DB_KEY_USERNAME] != loggedUsername)
              .toList()
          ..conversationType = ChatConversationType.values
              .byName(docData[DB_KEY_CONVERSATION_TYPE])
          ..data = docData;

        if (docData[DB_KEY_LAST_MESSAGE_TIME] != null &&
            docData[DB_KEY_LAST_MESSAGE_TIME] != 0) {
          QuerySnapshot query = await _col
              .doc(d.id)
              .collection(C_MESSAGES)
              .where(DB_KEY_TIMESTAMP, isGreaterThan: timestamp)
              .get();
          conversation.unread = query.size;

          //Last message
          DocumentSnapshot docLastMessage = await _col
              .doc(d.id)
              .collection(C_MESSAGES)
              .doc('${docData[DB_KEY_LAST_MESSAGE_TIME]}')
              .get();
          
          Map<String, dynamic> dataC = {};
          if (docLastMessage.data() != null) {
            dataC = docLastMessage.data() as Map<String, dynamic>;
            dataC[DB_KEY_CONTENT] = ChatHelper.decrypt(
                conversation.data[DB_KEY_ENCRYPT_KEY], dataC[DB_KEY_CONTENT]);
          }
          conversation.lastMessage = dataC;

          cons.add(conversation);
        } else {
          conversation.unread = 0;
          conversation.isEmptyMessages = true;
          conversation.lastMessage = {};
          cons.add(conversation);
        }

        List<ChatListConversation> conversations =
            List<ChatListConversation>.from(state.conversations
                    .where((e) => e.conversationId != d.id)) +
                [conversation];
        conversations.sort((a, b) => a.timestamp > b.timestamp ? 1 : -1);
        add(ChatUpdateListConversationEvent(conversations));
      }));
    }
  }

  void _getConnectionStatusUsers(List requestUsers) async {
    List<String> users =
        requestUsers.map((e) => e[DB_KEY_USERNAME].toString()).toList();
    Get.find<ConnectionStatusBloc>().add(ConnectionStatusInitEvent(users));
  }

  Future<Map> _getUsersByKeys(requestUsers) async {
    Map usersByKeys = state.usersByKeys;
    for (var e in requestUsers) {
      if (state.usersByKeys[e[DB_KEY_USERNAME]] == null) {
        Map<String, dynamic>? data = (await FirebaseFirestore.instance
                .collection(e[DB_KEY_USER_TYPE].endsWith('s')
                    ? e[DB_KEY_USER_TYPE]
                    : '${e[DB_KEY_USER_TYPE]}s')
                .doc(e[DB_KEY_USERNAME])
                .get())
            .data();
        usersByKeys[e[DB_KEY_USERNAME]] = data;
      }
    }
    return usersByKeys;
  }
}
